#include <stdlib.h>
#include <stdio.h>
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#include <math.h>

/* Meschach library */
#include "meschach/matrix.h"
#include "meschach/matrix2.h"

#define N 3 /* Nombre de morceaux du bras */

/* Matrices et vecteurs */
MAT *Jacobien;
VEC *Theta;
VEC *Longueurs;
MAT *Pivots;

int width, height;
int xmin, ymin;
float beginPoint[2]; 
float pos = 1;

void myinit();
void display();

void drawObject()
{
  int i;

  glPointSize(8.0);
  glLineWidth(5.0);

  glPushMatrix();
  glTranslatef(0, -0.9, 0);
  /* Le sol */
  glColor3f(1.0, 0.0, 0.0);
  glBegin(GL_LINES);
  glVertex2f(-1.0, 0.0);
  glVertex2f(1.0, 0.0);
  glEnd();

  glRotatef(-90.0, 0, 0, 1);
  glBegin(GL_POINTS);
  glVertex2f(0.0, 0.0);
  glEnd();
  
  for (i = 0; i < N; i++) {
    glColor3f(0.0, 1.0, 0.0);
    glRotatef((180.0f/M_PI)*Theta->ve[i], 0, 0, 1);
    glBegin(GL_LINES);
    glVertex2f(0.0, 0.0);
    glVertex2f(0.0, Longueurs->ve[i]);
    glEnd();

    glColor3f(1.0, 0.0, 0.0);
    glTranslatef(0.0, Longueurs->ve[i], 0.0);
    glBegin(GL_POINTS);
    glVertex2f(0.0, 0.0);
    glEnd();
  }
  glPopMatrix();
}

void computePivotCoordinates()
{
  int i;
  float angleCourant = 0;

  Pivots->me[0][0] = 0.0;
  Pivots->me[0][1] = 0.0;

  for (i = 1; i <= N; i++) {
    angleCourant += Theta->ve[i-1];
    Pivots->me[i][0] = Pivots->me[i-1][0] + Longueurs->ve[i-1]*cos(angleCourant);
    Pivots->me[i][1] = Pivots->me[i-1][1] + Longueurs->ve[i-1]*sin(angleCourant);
  }
}

float norm(VEC* E)
{
  return sqrt(E->ve[0]*E->ve[0] + E->ve[1]*E->ve[1]);
}

void computeCoordinates(float xp, float yp)
{
  MAT *U;
  MAT *Up;
  MAT *V;
  VEC* diagonale;
  MAT* diag;
  VEC* lambda;

  MAT* Maux;
  MAT* Maux2;
  MAT* JacInverse;
  int i, j;
  VEC* E;
  float normE;
  float epsilon = 1e-2;
  float lMax;
  int k, kmax = 50;

  U = m_get(2,2);
  V = m_get(N,N);
  E = v_get(2);
  diagonale = v_get(2);
  Maux = m_get(N,2);
  JacInverse = m_get(N,2);
  lambda = v_get(N);


  /* 1. Calcul vecteur diffrence de position */
  computePivotCoordinates();
  E->ve[0] = xp - Pivots->me[N][0];
  E->ve[1] = yp - Pivots->me[N][1];
  normE = norm(E);
  for (k = 0; k < kmax && (normE > epsilon) ; k++) {
    /* 2. Calcul du Jacobien */
    for (i = 0; i < N; i++) {
      Jacobien->me[0][i] = Pivots->me[i][1] - Pivots->me[N][1];
      Jacobien->me[1][i] = Pivots->me[N][0] - Pivots->me[i][0];
      /* printf("%d : %f %f \n", i, Jacobien->me[0][i], Jacobien->me[1][i] ); */
    }
    printf("Norme = %f \n", normE);
    fflush(stdout);
    /* 3. Dcomposition SVD du Jacobien */
    diagonale = svd(Jacobien, U, V, diagonale);

    /* 4. Inversion du Jacobien */
    diag = m_get(N, 2);
    for (i = 0; i < 2; i++) {
      if (diagonale->ve[i] > 0.001*diagonale->ve[0]) 
	diag->me[i][i] = 1.0f/diagonale->ve[i];
    }
    Maux = m_mlt(V, diag, Maux);   /* N.N * N.2 = N.2 */
    JacInverse = mmtr_mlt(Maux, U, JacInverse);  /* N.2 * (2.2)T = N.2 */

    /* 5. Application du pseudo-inverse du Jacobien */
    lambda = mv_mlt(JacInverse, E, lambda);
    printf("Lambda =");
    for  (i = 0; i < N; i++) 
      printf("%f ", lambda[i]);
    printf("\n");
    lMax = 0;
    for (i = 0; i < N; i++) 
      if (fabs(lambda->ve[i]) > lMax) lMax = fabs(lambda->ve[i]);
    if (lMax > M_PI/90.0f) {
      for (i = 0; i < N; i++) 
	lambda->ve[i] *= M_PI/(lMax * 90.0f);
      printf("Corrige =");
      for  (i = 0; i < N; i++) 
	printf("%f ", lambda[i]);
      printf("\n");
    }
    /* Addition du dplacement calcul: */
    Theta = v_add(Theta, lambda, Theta);
    printf("Theta =");
    for  (i = 0; i < N; i++) 
      printf("%f ", Theta->ve[i]);
    printf("\n");
    lMax = 0;
    computePivotCoordinates();
    display();
    E->ve[0] = xp - Pivots->me[N][0];
    E->ve[1] = yp - Pivots->me[N][1];
    normE = norm(E);
  }
}

void
mouse(int button, int state, int x, int y)
{
  float d;
  float cos_t;
  float xp, yp;
  
  if (state == GLUT_DOWN) {
    xp = -1 + (2.0*(x-xmin))/(float)height;
    yp = 1.9 - (2.0*(y-ymin))/(float)height;
    if (yp < 0) yp = 0;

    /* Calcul des coordonnees */
    computeCoordinates(xp, yp);
  } 
  glutPostRedisplay();
}

void
motion(int x, int y)
{
  float d;
  float cos_t;
  float xp, yp;
  
  xp = -1 + (2.0*(x-xmin))/(float)height;
  yp = 1.9 - (2.0*(y-ymin))/(float)height;
  if (yp < 0) yp = 0;

  /* Calcul des coordonnees */
  computeCoordinates(xp, yp);

  glutPostRedisplay();
}

void printHelp()
{
  printf("A small program to show N-link Inverse Kinematics\n\n");
}

void parsekey(unsigned char key, int x, int y)
{
  switch (key)
    {
    case 27: exit(0);
    case ' ':
      myinit();
      break;
    case 'h':
      printHelp();
      break;
    default:
      return;
    }
  glutPostRedisplay();
}

void parsekey_special(int key, int x, int y)
{
  switch (key)
    {
    case GLUT_KEY_UP:	
      break;
    case GLUT_KEY_DOWN:	
      break;
    case GLUT_KEY_RIGHT: 
      break;
    case GLUT_KEY_LEFT:	 
      break;
    default:
      return;
    }
  glutPostRedisplay();
}

void myReshape(int w, int h)
{
  xmin = 0;
  ymin = 0;
  if (w > h) {
    xmin = (w-h)/2;
    w = h;
  } else {
    ymin = (h-w)/2;
    h = w;
  }
  width = w;
  height = h;
  glViewport(xmin, ymin, w, h);
  glutPostRedisplay();
}

void display(void)
{
  glPushMatrix();
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

  drawObject();

  glPopMatrix();
  glutSwapBuffers(); 
}

void myinit(void)
{
  int i;

  glLoadIdentity();
  glClearColor(0.0, 0.0, 0.6, 1.0);

  Jacobien = m_get(2, N);
  Theta = v_get(N);
  Longueurs = v_get(N);
  Pivots = m_get(N+1, 2);
  for (i = 0; i < N; i++) {
    Longueurs->ve[i] = 1.0f/N;
    Theta->ve[i] = 0.0;
  }
}

int main(int argc, char *argv[]) 
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DEPTH | GLUT_RGB | GLUT_DOUBLE | GLUT_MULTISAMPLE |
		      GLUT_STENCIL);
  glutCreateWindow("Inverse Kinematics, N links");
  glutDisplayFunc(display);
  glutInitWindowSize(1200, 1200);
  glutInitWindowPosition(200, 0);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutKeyboardFunc(parsekey);
  glutSpecialFunc(parsekey_special);
  glutReshapeFunc(myReshape);
  myinit();
  printHelp();
    
  glutSwapBuffers();
  glutMainLoop();
  return 0;
}
