#include <stdlib.h>
#include <stdio.h>
#ifdef __APPLE__
#include <GLUT/glut.h>
#else
#include <GL/glut.h>
#endif
#include <math.h>

#define N 10 /* Nombre de morceaux du bras */

/* Matrices et vecteurs */
float **Jacobien;
float Theta[N];
float Longueurs[N];
float Pivots[N+1][2];

int width, height;
int xmin, ymin;
float beginPoint[2]; 
float pos = 1;

void myinit();
void display();
extern void svdcmp(float **a, int m, int n, float w[], float **v);

void drawObject()
{
  int i;

  glPointSize(8.0);
  glLineWidth(5.0);

  glPushMatrix();
  glTranslatef(0, -0.9, 0);
  /* Le sol */
  glColor3f(1.0, 0.0, 0.0);
  glBegin(GL_LINES);
  glVertex2f(-1.0, 0.0);
  glVertex2f(1.0, 0.0);
  glEnd();

  glRotatef(-90.0, 0, 0, 1);
  glBegin(GL_POINTS);
  glVertex2f(0.0, 0.0);
  glEnd();
  
  for (i = 0; i < N; i++) {
    glColor3f(0.0, 1.0, 0.0);
    glRotatef((180.0f/M_PI)*Theta[i], 0, 0, 1);
    glBegin(GL_LINES);
    glVertex2f(0.0, 0.0);
    glVertex2f(0.0, Longueurs[i]);
    glEnd();

    glColor3f(1.0, 0.0, 0.0);
    glTranslatef(0.0, Longueurs[i], 0.0);
    glBegin(GL_POINTS);
    glVertex2f(0.0, 0.0);
    glEnd();
  }
  glPopMatrix();
}

void computePivotCoordinates()
{
  int i;
  float angleCourant = 0;

  Pivots[0][0] = 0.0;
  Pivots[0][1] = 0.0;

  for (i = 1; i <= N; i++) {
    angleCourant += Theta[i-1];
    Pivots[i][0] = Pivots[i-1][0] + Longueurs[i-1]*cos(angleCourant);
    Pivots[i][1] = Pivots[i-1][1] + Longueurs[i-1]*sin(angleCourant);
  }
}

float norm(float E[2])
{
  return sqrt(E[0]*E[0] + E[1]*E[1]);
}

void computeCoordinates(float xp, float yp)
{
  float JacInverse[N][2];
  float** V;
  float diagonale[N+1];
  float lambda[N];

  int i, j, l;
  float E[2];
  float normE;
  float epsilon = 1e-2;
  float lMax;
  int k, kmax = 50;

  V = malloc((N+1)*sizeof(float*));
  for (i=0; i < (N+1); i++)
    V[i] = malloc((N+1)*sizeof(float));
  /* 1. Calcul vecteur diffrence de position */
  computePivotCoordinates();
  E[0] = xp - Pivots[N][0];
  E[1] = yp - Pivots[N][1];
  normE = norm(E);
  for (k = 0; k < kmax && (normE > epsilon) ; k++) {
    /* 2. Calcul du Jacobien */
    for (i = 1; i <= N; i++) {
      Jacobien[1][i] = Pivots[i-1][1] - Pivots[N][1];
      Jacobien[2][i] = Pivots[N][0] - Pivots[i-1][0];
    }
    /* 3. Dcomposition SVD du Jacobien */
    svdcmp(Jacobien, 2, N, diagonale, V);
    /* Jacobien : 2.N, diagonale : N, V : N.N, U : 2.N
       verif : 2.N * N.N * N.N = 2.N */
    /* 4. Inversion du Jacobien */
    lMax = diagonale[1]*0.1;
    for (i = 1; i <= N; i++) {
      if (diagonale[i] > lMax) 
	diagonale[i] = 1.0f/diagonale[i];
      else diagonale[i] = 0.0f;
    }

    /* Calcul du pseudo-inverse : */
    /* JacInverse = V * diag * t(Jacobien) */
    for (i=0; i < N; i++) {
      for (j = 0; j < 2; j++) {
	JacInverse[i][j] = 0;
	for (l = 0; l < N; l++) 
	  JacInverse[i][j] += V[i+1][l+1]*diagonale[l+1]*Jacobien[j+1][l+1];
      }
    }
    /* 5. Application du pseudo-inverse du Jacobien */
    /* lambda = JacInverse * E */
    for (i = 0; i < N; i++) {
      lambda[i] = JacInverse[i][0]*E[0] + JacInverse[i][1]*E[1];
    }
    lMax = 0;
    for (i = 0; i < N; i++) {
      if (fabs(lambda[i]) > lMax) lMax = fabs(lambda[i]);
    }
    if (lMax > M_PI/90.0f) {
      for (i = 0; i < N; i++) {
	lambda[i] *= M_PI/(lMax * 90.0f);
      }
    }
    /* Addition du dplacement calcul: */    
    /* Theta = Theta + lambda  */
    for (i = 0; i < N; i++) {
      Theta[i] += lambda[i];
    }
    computePivotCoordinates();
    display();
    E[0] = xp - Pivots[N][0];
    E[1] = yp - Pivots[N][1];
    normE = norm(E);
  }
}

void
mouse(int button, int state, int x, int y)
{
  float d;
  float cos_t;
  float xp, yp;
  
  if (state == GLUT_DOWN) {
    xp = -1 + (2.0*(x-xmin))/(float)height;
    yp = 1.9 - (2.0*(y-ymin))/(float)height;

    /* Calcul des coordonnees */
    computeCoordinates(xp, yp);
  } 
  glutPostRedisplay();
}

void
motion(int x, int y)
{
  float d;
  float cos_t;
  float xp, yp;
  
  xp = -1 + (2.0*(x-xmin))/(float)height;
  yp = 1.9 - (2.0*(y-ymin))/(float)height;

  /* Calcul des coordonnees */
  computeCoordinates(xp, yp);

  glutPostRedisplay();
}

void printHelp()
{
  printf("A small program to show N-link Inverse Kinematics\n\n");
}

void parsekey(unsigned char key, int x, int y)
{
  switch (key)
    {
    case 27: exit(0);
    case ' ':
      myinit();
      break;
    case 'h':
      printHelp();
      break;
    default:
      return;
    }
  glutPostRedisplay();
}

void parsekey_special(int key, int x, int y)
{
  switch (key)
    {
    case GLUT_KEY_UP:	
      break;
    case GLUT_KEY_DOWN:	
      break;
    case GLUT_KEY_RIGHT: 
      break;
    case GLUT_KEY_LEFT:	 
      break;
    default:
      return;
    }
  glutPostRedisplay();
}

void myReshape(int w, int h)
{
  xmin = 0;
  ymin = 0;
  if (w > h) {
    xmin = (w-h)/2;
    w = h;
  } else {
    ymin = (h-w)/2;
    h = w;
  }
  width = w;
  height = h;
  glViewport(xmin, ymin, w, h);
  glutPostRedisplay();
}

void display(void)
{
  glPushMatrix();
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);

  drawObject();

  glPopMatrix();

  glutSwapBuffers(); 
}

void myinit(void)
{
  int i;

  glLoadIdentity();
  glClearColor(0.0, 0.0, 0.6, 1.0);

  for (i = 0; i < N; i++) {
    Longueurs[i] = 1.5f/N;
    Theta[i] = 0.0f;
  }

  Jacobien = malloc(3*sizeof(float*));
  for (i=1; i <= 2; i++)
    Jacobien[i] = malloc((N+1)*sizeof(float));
}

int main(int argc, char *argv[]) 
{
  glutInit(&argc, argv);
  glutInitDisplayMode(GLUT_DEPTH | GLUT_RGB | GLUT_DOUBLE | GLUT_MULTISAMPLE |
		      GLUT_STENCIL);
  glutCreateWindow("Inverse Kinematics, N links");
  glutDisplayFunc(display);
  glutInitWindowSize(1200, 1200);
  glutInitWindowPosition(200, 0);
  glutMouseFunc(mouse);
  glutMotionFunc(motion);
  glutKeyboardFunc(parsekey);
  glutSpecialFunc(parsekey_special);
  glutReshapeFunc(myReshape);
  myinit();
  printHelp();
    
  glutSwapBuffers();
  glutMainLoop();
  return 0;
}
